#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""pid_4sc.py
@author: Louis Richard
"""


import numpy as np
import xarray as xr

from ..mms import rotate_tensor

from .c_4_grad import c_4_grad
from .avg_4sc import avg_4sc
from .trace import trace
from .ts_tensor_xyz import ts_tensor_xyz
from .ts_scalar import ts_scalar


def pid_4sc(r_mms, v_mms, p_mms, b_mms):
    """
    Compute Pi-D term using definition of [10]_ as :

    .. math::

        \\textrm{Pi}-\\textrm{D} = - \\mathbf{\\Pi}_{ij}\\mathbf{D}_{ij}

    with :math:`\\mathbf{\\Pi}_{ij}` the deviatoric part of the pressure
    tensor :

    .. math::

        \\mathbf{\\Pi}_{ij} = \\mathbf{P}_{ij} - p\\delta_{ij}

        p = \\frac{1}{3}\\mathbf{P}_{ii}


    and :math:`\\mathbf{D}_{ij}` the deviatoric part of the strain tensor :

    .. math::

        \\mathbf{D}_{ij} =
        \\frac{1}{2}\\left ( \\partial_i \\mathbf{u}_j + \\partial_j
        \\mathbf{u}_i\\right ) - \\frac{1}{3}\\theta\\delta_{ij}

        \\theta = \\nabla . \\mathbf{u}


    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the position of the 4 spacecraft.

    v_mms : list of xarray.DataArray
        Time series of the bulk velocities of the 4 spacecraft.

    p_mms : list of xarray.DataArray
        Time series of the pressure tensor of the 4 spacecraft.

    b_mms : list of xarray.DataArray
        Time series of the background magnetic field of the 4 spacecraft.

    Returns
    -------
    pid : xarray.DataArray
        Time series of the Pi-D.

    References
    ----------
    .. [10]  Yang, Y., Matthaeus, W. H., Parashar, T. N., Wu, P., Wan, M.,
            Shi, Y., et al. (2017). Energy transfer channels and turbulence
            cascade in Vlasov-Maxwell turbulence. Physical Review E, 95,
            061201. doi : https://doi.org/10.1103/PhysRevE.95.061201

    """

    # Compute pressure tensor and background magnetic field at the center of
    # mass of the tetrahedron
    p_xyz = avg_4sc(p_mms)
    b_xyz = avg_4sc(b_mms)

    # Compute divergence and gradient tensor of the bulk velocity. Yang's
    # notation
    theta = c_4_grad(r_mms, v_mms, "div")
    grad_u = c_4_grad(r_mms, v_mms, "grad")

    # Define identity tensor
    identity_3d = np.zeros((len(grad_u), 3, 3))
    np.einsum("jii->ji", identity_3d)[:] = 1

    # strain tensor
    eps_xyz = (grad_u.data + np.transpose(grad_u.data, [0, 2, 1])) / 2

    # Deviatoric part of the strain tensor
    d_xyz = eps_xyz - theta.data[:, np.newaxis, np.newaxis] * identity_3d / 2
    d_xyz = ts_tensor_xyz(v_mms[0].time.data, d_xyz)

    # Convert tensors to field aligned coordinates
    d_fac = rotate_tensor(d_xyz, "fac", b_xyz, "pp")
    p_fac = rotate_tensor(p_xyz, "fac", b_xyz, "pp")

    # Isotropic scalar pressure
    press = trace(p_fac) / 3

    # Deviatoric part of the pressure tensor
    pi_fac = p_fac.data - press.data[:, np.newaxis, np.newaxis] * identity_3d
    pi_fac = ts_tensor_xyz(v_mms[0].time.data, pi_fac)

    # Compute Pi-D
    pid = np.sum(np.sum(pi_fac.data * d_fac.data, axis=1), axis=1)
    pid = ts_scalar(v_mms[0].time.data, pid)

    # Flatten tensors
    d_flat = np.reshape(d_fac.data, [len(d_fac), 9])
    pi_flat = np.reshape(pi_fac.data, [len(pi_fac), 9])
    d_flat = d_flat[:, [0, 4, -1, 1, 2, 5]]
    pi_flat = pi_flat[:, [0, 4, -1, 1, 2, 5]]

    # Compute components of the double contraction sum
    d_coeff = xr.DataArray(d_flat * pi_flat,
                           coords=[p_xyz.time.data, np.arange(1, 7)],
                           dims=["time", "index"])

    return pid, d_coeff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import matplotlib.pyplot as plt

from pandas.plotting import register_matplotlib_converters
register_matplotlib_converters()


def colorbar(image, axis, pad=0.01):
    """Add colorbar to ax corresponding to im.

    Parameters
    ----------
    image : to fill
        to fill.

    axis : matplotlib.pyplot.subplotsaxes
        Axis of plot.

    pad : float, optional


    Returns
    -------
    cax : coloraxis

    """

    pos = axis.get_position()
    fig = plt.gcf()
    cax = fig.add_axes([pos.x0 + pos.width + pad, pos.y0, pad, pos.height])
    fig.colorbar(image, cax)

    return cax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""extend_tint.py
@author: Louis Richard
"""

from astropy.time import Time
from dateutil.parser import parse as date_parse


def extend_tint(tint, ext=None):
    """Extends time interval.

    Parameters
    ----------
    tint : list of str
        Reference time interval to extend.

    ext : list of float or list of int
        Number of seconds to extend time interval
        [left extend, right extend].

    Returns
    -------
    tint : list of str
        Extended time interval.

    Examples
    --------
    >>> from pyrfu import pyrf

    Time interval

    >>> tints = ["2015-10-30T05:15:42.000", "2015-10-30T05:15:54.000"]

    Load spacecraft position

    >>> tints_long = pyrf.extend_tint(tint, [-100, 100])

    """

    if ext is None:
        ext = [-60, 60]

    # Convert to unix format
    start_time = Time(date_parse(tint[0]), format="datetime").unix
    end_time = Time(date_parse(tint[1]), format="datetime").unix

    # extend interval
    start_time, end_time = [start_time + ext[0], end_time + ext[1]]

    # back to iso format
    start_time = Time(start_time, format="unix").iso
    end_time = Time(end_time, format="unix").iso

    tint_long = [start_time, end_time]

    return tint_long



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""geocentric_coordinate_transformation.py
@author: Louis Richard
"""

import os
import yaml
import numpy as np
import xarray as xr

from astropy.time import Time
from ..models import igrf

from .ts_vec_xyz import ts_vec_xyz


def triang(angle, axis):
    """angle in degrees ax=1 > X, ax=2 > y,ax=3 > z)

    Parameters
    ----------
    angle : ndarray
        Angles in degres.

    axis : int
        Index of the rotation axis 0, 1, 2

    Returns
    -------
    transf_mat : ndarray
        Matrix of rotation of the angle around the axis.

    """

    cos_angle = np.cos(np.deg2rad(angle))
    sin_angle = np.sin(np.deg2rad(angle))

    axes = list(filter(lambda x: x != axis, np.arange(3)))

    transf_mat = np.zeros((len(angle), 3, 3))
    transf_mat[:, axes[0], axes[0]] = cos_angle
    transf_mat[:, axes[1], axes[1]] = cos_angle
    transf_mat[:, axis, axis] = 1
    transf_mat[:, axes[0], axes[1]] = sin_angle
    transf_mat[:, axes[1], axes[0]] = -sin_angle

    return transf_mat


def dipole_direction_gse(time, flag: str = "dipole"):
    """Compute IGRF model dipole in GSE coordinates.

    Parameters
    ----------
    time : ndarray
        Times in unix format.

    flag : str
        Default is "dipole"


    Returns
    -------
    dipole_direction_gse_ : ndarray
        Direction of the IGRF dipole model in GSE coordinates. [t, x, y, z]
    """

    lambda_, phi = igrf(time, flag)

    cos_phi = np.cos(np.deg2rad(phi))
    dipole_direction_geo_ = np.stack([cos_phi * np.cos(np.deg2rad(lambda_)),
                                      cos_phi * np.sin(np.deg2rad(lambda_)),
                                      np.sin(np.deg2rad(phi))]).T

    dipole_direction_gse_ = geocentric_coordinate_transformation(np.hstack([
        time[:, None], dipole_direction_geo_]), "geo>gse")

    return dipole_direction_gse_


def transformation_matrix(t, tind, hapgood, *args):
    """Compute the transformation matrix associated to the selected
    coordinates system change.

    Parameters
    ----------
    t : ndarray
        Time in unix format.

    tind : list
        Transformation indices.

    hapgood : bool
        Flag to use Hapgoog

    args : tuple
        to fill.

    Returns
    -------
    transf_mat_out : ndarray
        Transformation matrix


    """
    t_zero, ut, d0_j2000, d_j2000, h_j2000, t_j2000 = args

    transf_mat_out = np.zeros((len(t), 3, 3))
    transf_mat_out[:, 0, 0] = np.ones(len(t))
    transf_mat_out[:, 1, 1] = np.ones(len(t))
    transf_mat_out[:, 2, 2] = np.ones(len(t))

    for j, t_num in enumerate(tind[::-1]):
        if t_num == 1 or t_num == -1:
            if hapgood:
                theta = 100.461 + 36000.770 * t_zero + 15.04107 * ut

            else:
                # Source: United States Naval Observatory, Astronomical
                # Applications Dept. http: // aa.usno.navy.mil / faq / docs
                # / GAST.php Last modified: 2011/06/14T14:04
                gmst = 6.697374558
                gmst += 0.06570982441908 * d0_j2000
                gmst += 1.00273790935 * h_j2000
                gmst += 0.000026 * t_j2000 ** 2

                gmst = gmst % 24  # Interval 0->24 hours
                theta = (360 / 24) * gmst  # Convert to degree.

            # invert if tInd = -1
            transf_mat = triang(theta * np.sign(t_num), 2)

        elif t_num == 2 or t_num == -2:
            if hapgood:
                eps = 23.439 - 0.013 * t_zero
                # Suns mean anomaly
                m_anom = 357.528 + 35999.050 * t_zero + 0.04107 * ut
                # Suns mean longitude
                m_long = 280.460 + 36000.772 * t_zero + 0.04107 * ut
                l_sun = m_long
                l_sun += (1.915 - 0.0048 * t_zero) * np.sin(np.deg2rad(m_anom))
                l_sun += 0.020 * np.sin(np.deg2rad(2 * m_anom))
            else:
                # Source: United States Naval Observatory, Astronomical
                # Applications Dept.
                # http://aa.usno.navy.mil/faq/docsSunApprox.php.
                # Last modified: 2012/11/06T14:12
                eps = 23.439 - 0.00000036 * d_j2000
                m_anom = 357.529 + 0.98560028 * d_j2000
                m_long = 280.459 + 0.98564736 * d_j2000
                l_sun = m_long
                l_sun += 1.915 * np.sin(np.deg2rad(m_anom))
                l_sun += 0.020 * np.sin(np.deg2rad(2 * m_anom))

            transf_mat = np.matmul(triang(l_sun, 2), triang(eps, 0))
            if t_num == -2:
                transf_mat = np.transpose(transf_mat, [0, 2, 1])

        elif t_num == 3 or t_num == -3:
            dipole_direction_gse_ = dipole_direction_gse(t, "dipole")
            y_e = dipole_direction_gse_[:, 2]  # 1st col is time
            z_e = dipole_direction_gse_[:, 3]
            psi = np.rad2deg(np.arctan(y_e / z_e))

            transf_mat = triang(-psi * np.sign(t_num), 0)  # inverse if -3

        elif t_num == 4 or t_num == -4:
            dipole_direction_gse_ = dipole_direction_gse(t, "dipole")

            mu = np.arctan(dipole_direction_gse_[:, 1]
                           / np.sqrt(np.sum(dipole_direction_gse_[:, 2:] ** 2,
                                            axis=1)))
            mu = np.rad2deg(mu)

            transf_mat = triang(-mu * np.sign(t_num), 1)

        elif t_num == 5 or t_num == -5:
            lambda_, phi = igrf(t, "dipole")

            transf_mat = np.matmul(triang(phi - 90, 1), triang(lambda_, 2))
            if t_num == -5:
                transf_mat = np.transpose(transf_mat, [0, 2, 1])

        else:
            raise ValueError

        if j == len(tind):
            transf_mat_out = transf_mat
        else:
            transf_mat_out = np.matmul(transf_mat, transf_mat_out)

    return transf_mat_out


def geocentric_coordinate_transformation(inp, flag, hapgood: bool = True):
    """IRF.GEOCENTRIC_COORDINATE_TRANSFORMATION coordinate transformation
    GE0/GEI/GSE/GSM/SM/MAG

    Parameters
    ----------
    inp : xarray.DataArray or ndarray
        Time series of the input field.

    flag : str
        Coordinates transformation "{coord1}>{coord2}", where coord1 and
        coord2 can be geo/gei/gse/gsm/sm/mag.

    hapgood : bool
        Indicator if original Hapgood sources should be used for angle
        computations or if updated USNO-AA sources should be used.
        Default = true, meaning original Hapgood sources.


    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu.pyrf import geocentric_coordinate_transformation

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field in GSE coordinates

    >>> b_gse = get_data("b_gse_fgm_srvy_l2", tint, mms_id)

    Transform to GSM assuming that the original coordinates system is part
    of the inp metadata

    >>> b_gsm = geocentric_coordinate_transformation(b_gse, 'GSM')

    If the original coordinates is not in the meta

    >>> b_gsm = geocentric_coordinate_transformation(b_gse, 'GSE>GSM')

    Compute the dipole direction in GSE

    >>> dipole = geocentric_coordinate_transformation(b_gse.time,
    'dipoledirectiongse')


    References
    ----------
    .. [17]     Hapgood 1997 (corrected version of Hapgood 1992) Planet.Space
                Sci..Vol. 40, No. 5. pp. 71l - 717, 1992

    .. [18]     USNO - AA 2011 & 2012

    """

    if ">" in flag:
        ref_syst_in, ref_syst_out = flag.split(">")
    else:
        ref_syst_in, ref_syst_out = [None, flag.lower()]

    if isinstance(inp, xr.DataArray):
        if "COORDINATE_SYSTEM" in inp.attrs:
            ref_syst_internal = inp.attrs["COORDINATE_SYSTEM"].lower()
            ref_syst_internal = ref_syst_internal.split(">")[0]
        else:
            ref_syst_internal = None

        if ref_syst_in is None and ref_syst_internal is None:
            raise ValueError("input reference frame undefined")

        elif ref_syst_in is not None and ref_syst_internal is not None:
            message = "input ref. frame in variable and input flag differs"
            assert ref_syst_internal == ref_syst_in, message

        elif ref_syst_in is None and ref_syst_internal is not None:
            ref_syst_in = ref_syst_internal.lower()

        flag = f"{ref_syst_in}>{ref_syst_out}"

    if ref_syst_in == ref_syst_out:
        return inp

    # J2000 reference time
    j2000 = 946727930.8160001
    # j2000 = Time("J2000", format="jyear_str").unix

    if isinstance(inp, xr.DataArray):
        time = inp.time.data
        t = time.view("i8") * 1e-9

        #  Terrestial Time (seconds since J2000)
        tts = t - j2000
        inp_ts = inp
        inp = inp.data

    elif isinstance(inp, np.ndarray):
        time = Time(inp[:, 0], format="unix").datetime64
        t = inp[:, 0]
        #  Terrestial Time (seconds since J2000)
        tts = t - j2000
        inp_ts = None
        inp = inp[:, 1:]
    else:
        raise TypeError("invalid inpu")

    if hapgood:
        day_start_epoch = Time(time.astype("datetime64[D]"),
                               format="datetime64").unix
        mjd_ref_epoch = Time("2000-01-01T12:00:00", format="isot").unix

        # t_zero is time measured in Julian centuries from 2000-01-0112:00 UT
        # to the previous midnight
        t_zero = (day_start_epoch - mjd_ref_epoch)
        t_zero /= (3600 * 24 * 36525.0)

        hours = (time.astype('datetime64[h]')
                 - time.astype('datetime64[D]')).astype(float)
        minutes = (time.astype('datetime64[m]')
                   - time.astype('datetime64[h]')).astype(float)
        seconds = 1e-9 * (time.astype('datetime64[ns]')
                          - time.astype('datetime64[m]')).astype(float)
        ut = hours + minutes / 60 + seconds / 3600

        args_trans_mat = (t_zero, ut, None, None, None, None)

    else:
        # Julian date(of req.time) from J2000
        d_j2000 = tts / 86400

        # Julian date(of preceeding midnight of req.time) from J2000
        d0_j2000 = np.floor(tts / 86400) - 0.5

        # Julian centuries(of req.time) since J2000
        t_j2000 = d_j2000 / 36525

        # Hours in the of req.time(since midnight).
        h_j2000 = 24 * (d_j2000 - d0_j2000)

        args_trans_mat = (None, None, d0_j2000, d_j2000, h_j2000, t_j2000)

    if ">" in flag:
        root_path = os.path.dirname(os.path.abspath(__file__))
        file_name = "transformation_indices.yml"

        with open(os.sep.join([root_path, file_name])) as file:
            transformation_dict = yaml.load(file, Loader=yaml.FullLoader)

        tind = transformation_dict[flag]

    elif flag == "dipoledirectiongse":
        out_data = dipole_direction_gse(t)
        return ts_vec_xyz(inp.time.data, out_data)

    else:
        raise ValueError(f"Transformation {flag} is unknown!")

    transf_mat = transformation_matrix(t, tind, hapgood, *args_trans_mat)

    if inp.ndim == 2:
        out = np.einsum('kji,ki->kj', transf_mat, inp)
    elif inp.ndim == 1:
        out = transf_mat
    else:
        raise ValueError

    if inp_ts is not None:
        out_data = out
        out = inp_ts.copy()
        out.data = out_data
        out.attrs["COORDINATE_SYSTEM"] = ref_syst_out.upper()

    else:
        out = np.hstack([t[:, None], out])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""sph2cart.py
@author: Louis Richard
"""

import numpy as np


def sph2cart(azimuth, elevation, r):
    """Transform spherical to cartesian coordinates

    Parameters
    ----------
    azimuth : float or ndarray
        Azimuthal angle (phi)

    elevation : float or ndarray
        Elevation angle (theta)

    r : float or ndarray
        Radius

    Returns
    -------
    x : float or ndarray
        Cartesian x-axis coordinates.

    y : float or ndarray
        Cartesian y-axis coordinates.

    z : float or ndarray
        Cartesian z-axis coordinates

    """
    x = r * np.cos(elevation) * np.cos(azimuth)
    y = r * np.cos(elevation) * np.sin(azimuth)
    z = r * np.sin(elevation)

    return x, y, z


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.
import warnings
import xarray as xr
import matplotlib.pyplot as plt

from ..pyrf import histogram2d
from . import plot_spectr


def pl_scatter_matrix(inp1=None, inp2=None, pdf=False, cmap="jet"):
    """Produces a scatter plot of each components of field inp1 with respect to every component
    of field inp2. If pdf is set to True, the scatter plot becomes a 2d histogram.

    Parameters
    ----------
    inp1 : xarray.DataArray
        First time series (x-axis).

    inp2 : xarray.DataArray
        Second time series (y-axis).

    m : str, optional
        Marker type. Default is "+". Not used if pdf is True

    pdf : bool
        Flag to plot the 2d histogram. If False the figure is a scatter plot. If True the figure
        is a 2d histogram.

    cmap : str
        Colormap. Default : "jet"

    Returns
    -------
    fig : matplotlib.pyplot.figure
        to fill.

    axs : matplotlib.pyplot.axes
        to fill.

    caxs : matplotlib.pyplot.colorbar
        Only if pdf is True

    """

    if inp1 is None:
        raise ValueError("pl_scatter_matrix requires at least one argument")

    if inp2 is None:
        inp2 = inp1
        warnings.warn("inp2 is empty assuming that inp2=inp1", UserWarning)

    if not isinstance(inp1, xr.DataArray) or not isinstance(inp2, xr.DataArray):
        raise TypeError("Inputs must be DataArrays")

    if not pdf:
        fig, axs = plt.subplots(3, 3, sharex="all", sharey="all", figsize=(16, 9))
        fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9, hspace=0.05, wspace=0.05)

        for i in range(3):
            for j in range(3):
                axs[j, i].scatter(inp1[:, i].data, inp2[:, j].data, marker="+")

        out = (fig, axs)
    else:
        fig, axs = plt.subplots(3, 3, sharex="all", sharey="all", figsize=(16, 9))
        fig.subplots_adjust(left=0.1, right=0.9, bottom=0.1, top=0.9, hspace=0.05, wspace=0.3)

        hist = [[None]*3]*3

        caxs = [[None]*3]*3

        for i in range(3):
            for j in range(3):
                hist[j][i] = histogram2d(inp1[:, i], inp2[:, j])
                axs[j, i], caxs[j][i] = plot_spectr(axs[j, i], hist[j][i], cmap=cmap, cscale="log")
                axs[j, i].grid()

        out = (fig, axs, caxs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

"""resample.py
@author: Louis Richard
"""

import bisect
import warnings
import numpy as np
import xarray as xr

from scipy import interpolate


def guess_sampling_frequency(ref_time):
    """Compute sampling frequency of the time line.

    Parameters
    ----------
    ref_time : np.ndarray
        Time array

    Returns
    -------
    sfy : float
        Sampling frequency.

    """

    n_data = len(ref_time)

    sfy1 = 1 / (ref_time[1] - ref_time[0])
    sfy = None
    not_found = True

    if n_data == 2:
        sfy = sfy1
        not_found = False

    cur, max_try = [2, 10]

    while not_found and cur <= n_data and cur - 3 < max_try:
        sfy = 1 / (ref_time[cur] - ref_time[cur - 1])

        if np.absolute(sfy - sfy1) < sfy * .001:
            not_found = False

            sfy = (sfy + sfy1) / 2
            break

        sfy = sfy1
        cur += 1

    if not_found:
        raise RuntimeError(
            "Cannot guess sampling frequency. Tried {:d} times".format(
                max_try))

    return sfy


def average(inp_time, inp_data, ref_time, thresh, dt2):
    """Resample inp_data to timeline of ref_time, using half-window of dt2.
    Points above std*tresh are excluded. thresh=0 turns off this option.

    Parameters
    ----------
    inp_time : np.ndarray
        Original timeline of the input.

    inp_data : np.ndarray
        Time series data.

    ref_time : np.ndarray
        Reference timeline.

    thresh : float
        Threshold

    dt2 : float
        Size of the half-window.


    Returns
    -------
    out_data : np.ndarray
        Resampled data.

    """

    try:
        out_data = np.zeros((len(ref_time), inp_data.shape[1]))
    except IndexError:
        inp_data = inp_data[:, None]
        out_data = np.zeros((len(ref_time), inp_data.shape[1]))

    for i, ref_t in enumerate(ref_time):
        idx_l = bisect.bisect_left(inp_time, ref_t - dt2)
        idx_r = bisect.bisect_right(inp_time, ref_t + dt2)

        idx = np.arange(idx_l, idx_r)

        if idx.size == 0:
            out_data[i, ...] = np.nan
        else:
            if thresh:
                std_ = np.std(inp_data[idx, ...], axis=0)
                mean_ = np.mean(inp_data[idx, ...], axis=0)

                assert any(np.isnan(std_))

                for j, stdd in enumerate(std_):
                    if not np.isnan(stdd):
                        idx_r = bisect.bisect_right(
                            inp_data[idx, j + 1] - mean_[j], thresh * stdd)
                        if idx_r:
                            out_data[i, j + 1] = np.mean(
                                inp_data[idx[idx_r], j + 1], axis=0)
                        else:
                            out_data[i, j + 1] = np.nan
                    else:
                        out_data[i, ...] = np.nan

            else:
                out_data[i, ...] = np.mean(inp_data[idx, ...], axis=0)

    if out_data.shape[1] == 1:
        out_data = out_data[:, 0]

    return out_data


def resample(inp, ref, method="", f_s=None, window=None, thresh=0):
    """
    Resample inp to the time line of ref. If sampling of X is more than two
    times higher than Y, we average X, otherwise we interpolate X.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to resample.

    ref : xarray.DataArray
        Reference time line.

    method : str
        Method of interpolation "spline", "linear" etc.
        (default "linear") if method is given then interpolate
        independent of sampling.

    f_s : float
        Sampling frequency of the Y signal, 1/window.

    window : int or float or np.ndarray
        Length of the averaging window, 1/fsample.

    thresh : float
        Points above STD*THRESH are disregarded for averaging

    Returns
    -------
    out : xarray.DataArray
        Resampled input to the reference time line using the selected method.


    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)

    Resample magnetic field to electric field sampling

    >>> b_xyz = pyrf.resample(b_xyz, e_xyz)

    """

    flag_do = "check"

    if method:
        flag_do = "interpolation"

    if f_s is not None:
        sfy = f_s
    elif window is not None:
        sfy = 1 / window
    else:
        sfy = None

    inp_time = inp.time.data.view("i8") * 1e-9
    ref_time = ref.time.data.view("i8") * 1e-9

    if flag_do == "check":
        if len(ref_time) > 1:
            if not sfy:
                sfy = guess_sampling_frequency(ref_time)

            if len(inp_time) / (inp_time[-1] - inp_time[0]) > 2 * sfy:
                flag_do = "average"
                warnings.warn("Using averages in resample", UserWarning)
            else:
                flag_do = "interpolation"
        else:
            flag_do = "interpolation"

    assert flag_do in ["average", "interpolation"]

    if flag_do == "average":
        assert not method, "cannot mix interpolation and averaging flags"

        if not sfy:
            sfy = guess_sampling_frequency(ref_time)

        out_data = average(inp_time, inp.data, ref_time, thresh, .5 / sfy)

    else:
        if not method:
            method = "linear"

        # If time series agree, no interpolation is necessary.
        if len(inp_time) == len(ref_time) and all(inp_time == ref_time):
            out_data = inp.data.copy()
            coord = [ref.coords["time"].data]

            if len(inp.coords) > 1:
                for k in inp.dims[1:]:
                    coord.append(inp.coords[k].data)

            out = xr.DataArray(out_data, coords=coord, dims=inp.dims,
                               attrs=inp.attrs)

            return out

        tck = interpolate.interp1d(inp_time, inp.data, kind=method, axis=0,
                                   fill_value="extrapolate")
        out_data = tck(ref_time)

    coord = [ref.coords["time"]]

    if len(inp.coords) > 1:
        for k in inp.dims[1:]:
            coord.append(inp.coords[k].data)

    out = xr.DataArray(out_data, coords=coord, dims=inp.dims, attrs=inp.attrs)

    return out



# MIT License
#
# Copyright (c) 2020 Louis Richard
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.

import os
import shutil
import requests

from PIL import Image
from dateutil import parser as date_parser


def current_location(time, view="xy", path=".", show=False):
    """Download MMS location plots from MMS SDC.

    Parameters
    ----------
    time : str
        Time at which the location is wanted.

    view : str
        View mode xy or xz.

    path : str
        Path where to save images.

    show : bool
        Flag to show orbit plot.

    """

    # URL of the MMS Science Data Center (SDC)
    sdc_url = "https://lasp.colorado.edu/mms/sdc/public/data/sdc"

    # Date string format of time (needs only up to hours)
    str_time = date_parser.parse(time).strftime("%Y%m%d%H0000")

    # Name of the directory and the image corresponding to the view mode
    if view.lower() == "xy":
        dir_path = "mms_orbit_plots"
        img_name = "mms_orbit_plot_{}.png".format(str_time)
    elif view.lower() == "xz":
        dir_path = "mms_orbit_plots"
        img_name = "mms_orbit_plot_{}_{}.png".format(view, str_time)
    else:
        raise ValueError("Invalid view mode")

    # Create directory in the target path
    os.makedirs(os.path.join(path, dir_path), exist_ok=True)

    # Full url and path of the orbit plot
    img_url, img_path = [os.path.join(root, dir_path, img_name) for root in [sdc_url, path]]

    # Download image
    response = requests.get(img_url, stream=True)

    with open(img_path, "wb") as file:
        shutil.copyfileobj(response.raw, file)

    del response

    # Show the plot
    if show:
        img = Image.open(img_path)
        img.show()
